from typing import List, Callable, Any
from edulint.config.arg import Arg
from edulint.linting.problem import ProblemJson, Problem
from edulint.linting.process_handler import ProcessHandler
from edulint.linting.tweakers import get_tweakers, Tweakers
from edulint.config.config import Config
from edulint.linters import Linters
import sys
import json
import pathlib


def flake8_to_problem(raw: ProblemJson) -> Problem:
    assert isinstance(raw["filename"], str), f'got {type(raw["filename"])} for filename'
    assert isinstance(raw["line_number"], int), f'got {type(raw["line_number"])} for line_number'
    assert isinstance(raw["column_number"], int), f'got {type(raw["column_number"])} for column_number'
    assert isinstance(raw["code"], str), f'got {type(raw["code"])} for code'
    assert isinstance(raw["text"], str), f'got {type(raw["text"])} for text'

    return Problem(
        Linters.FLAKE8,
        raw["filename"],
        raw["line_number"],
        raw["column_number"],
        raw["code"],
        raw["text"]
    )


def pylint_to_problem(raw: ProblemJson) -> Problem:
    assert isinstance(raw["path"], str), f'got {type(raw["path"])} for path'
    assert isinstance(raw["line"], int), f'got {type(raw["line"])} for line'
    assert isinstance(raw["column"], int), f'got {type(raw["column"])} for column'
    assert isinstance(raw["message-id"], str), f'got {type(raw["message-id"])} for message-id'
    assert isinstance(raw["message"], str), f'got {type(raw["message"])} for message'
    assert isinstance(raw["endLine"], int) or raw["endLine"] is None, f'got {type(raw["endLine"])} for endLine'
    assert isinstance(raw["endColumn"], int) or raw["endColumn"] is None, f'got {type(raw["endColumn"])} for endColumn'

    return Problem(
        Linters.PYLINT,
        raw["path"],
        raw["line"],
        raw["column"],
        raw["message-id"],
        raw["message"],
        raw["endLine"],
        raw["endColumn"]
    )


def lint_any(
        linter: Linters, filename: str, args: List[str], config: List[str],
        result_getter: Callable[[Any], Any],
        out_to_problem: Callable[[ProblemJson], Problem]) -> List[Problem]:
    command = [sys.executable, "-m", str(linter)] + args + config + [filename]
    return_code, outs, errs = ProcessHandler.run(command, timeout=10)
    if (linter == Linters.FLAKE8 and return_code not in (0, 1)) or (linter == Linters.PYLINT and return_code == 32):
        print(errs, file=sys.stderr, end="")
        print(f"edulint: {command[2]} exited with {return_code}", file=sys.stderr)
        exit(return_code)
    if not outs:
        return []
    result = result_getter(json.loads(outs))
    return list(map(out_to_problem, result))


def lint_flake8(filename: str, config: Config) -> List[Problem]:
    flake8_args = ["--format=json"]
    return lint_any(
        Linters.FLAKE8, filename, flake8_args, config.others[Linters.FLAKE8],
        lambda r: r[filename],
        flake8_to_problem)


def lint_pylint(filename: str, config: Config) -> List[Problem]:
    cwd = pathlib.Path(__file__).parent.resolve()
    pylint_args = [f'--rcfile={cwd}/.pylintrc', "--output-format=json"]
    return lint_any(
        Linters.PYLINT, filename, pylint_args, config.others[Linters.PYLINT],
        lambda r: r, pylint_to_problem)


def apply_tweaks(problems: List[Problem], tweakers: Tweakers, args: List[Arg]) -> List[Problem]:
    result = []
    for problem in problems:
        tweaker = tweakers.get((problem.source, problem.code))
        if tweaker:
            if tweaker.should_keep(problem, [arg for arg in args if arg.option in tweaker.used_options]):
                problem.text = tweaker.get_reword(problem)
                result.append(problem)
        else:
            result.append(problem)
    return result


def lint(filename: str, config: Config) -> List[Problem]:
    result = lint_flake8(filename, config) + lint_pylint(filename, config)
    result = apply_tweaks(result, get_tweakers(), config.edulint)
    result.sort(key=lambda problem: (problem.line, problem.column))
    return result



from edulint.linters import Linters
from edulint.options import Option
from typing import Dict, List
from dataclasses import dataclass


@dataclass(frozen=True)
class Translation:
    to: Linters
    val: List[str]


CONFIG_TRANSLATIONS: Dict[Option, Translation] = {
    Option.ENHANCEMENT: Translation(
        Linters.PYLINT,
        ["--enable=no-self-use,superfluous-parens,consider-using-min-builtin,"
         "consider-using-max-builtin,consider-using-with,unspecified-encoding,"
         "unused-variable,unused-argument"]
    ),
    Option.PYTHON_SPEC: Translation(
        Linters.PYLINT,
        ["--enable=unidiomatic-typecheck,misplaced-format-function,"
         "unnecessary-lambda,protected-access,multiple-imports,"
         "wrong-import-position,consider-using-from-import,wildcard-import,"
         "reimported,consider-using-enumerate,consider-iterating-dictionary,"
         "consider-using-dict-items,consider-using-f-string,"
         "inconsistent-return-statements,consider-swap-variables,"
         "consider-using-join,consider-using-set-comprehension,"
         "unnecessary-comprehension,use-a-generator,use-list-literal,"
         "use-dict-literal,invalid-name"]
    ),
    Option.ALLOWED_ONECHAR_NAMES: Translation(
        Linters.PYLINT,
        ["--enable=invalid-name"]
    )
}


def get_config_translations() -> Dict[Option, Translation]:
    return CONFIG_TRANSLATIONS



from edulint.config.arg import Arg
from edulint.options import Option, TakesVal, OptionParse, get_option_parses, get_name_to_option
from edulint.linters import Linters
from edulint.config.config_translates import get_config_translations, Translation
from typing import Dict, List, Optional, Tuple, Any
import re
import sys
import shlex

ConfigDict = Dict[Linters, List[str]]


class Config:
    def __init__(self, edulint: Optional[List[Arg]] = None, others: Optional[ConfigDict] = None) -> None:
        self.edulint = edulint if edulint is not None else []
        others = others if others is not None else {}
        self.others: ConfigDict = {linter: others.get(linter, []) for linter in Linters}

    @staticmethod
    def get_val_from(args: List[Arg], option: Option) -> Optional[str]:
        for arg in args:
            if arg.option == option:
                return arg.val
        return None

    @staticmethod
    def has_opt_in(args: List[Arg], option: Option) -> bool:
        return Config.get_val_from(args, option) is not None

    def get_val(self, option: Option) -> Optional[str]:
        return Config.get_val_from(self.edulint, option)

    def has_opt(self, option: Option) -> bool:
        return Config.has_opt_in(self.edulint, option)

    def __str__(self) -> str:
        def to_str(linter: Linters, options: Any) -> str:
            return f"{linter}: {options}"
        return "{" \
            + ", ".join(to_str(linter, options) for linter, options in self.others.items() if options) \
            + ", " + to_str(Linters.EDULINT, self.edulint) \
            + "}"

    def __repr__(self) -> str:
        return str(self)


def extract_args(filename: str) -> List[str]:
    edulint_re = re.compile(r"\s*#[\s#]*edulint:\s*", re.IGNORECASE)

    result: List[str] = []
    with open(filename) as f:
        for i, line in enumerate(f):
            line = line.strip()
            edmatch = edulint_re.match(line)
            if not edmatch:
                continue

            raw_args = line[edmatch.end():]
            result.extend(shlex.split(raw_args))
    return result


def parse_args(args: List[str], option_parses: Dict[Option, OptionParse]) -> List[Arg]:
    name_to_option = get_name_to_option(option_parses)

    def get_name_val(arg: str) -> Tuple[str, Optional[str]]:
        if "=" in arg:
            assert arg.count("=") == 1
            name, val = arg.split("=")
            return name, val
        return arg, None

    result: List[Arg] = []
    for arg in args:
        name, val = get_name_val(arg)
        option = name_to_option.get(name)

        if option is None:
            print(f"edulint: unrecognized option {name}", file=sys.stderr)
        else:
            option_parse = option_parses[option]
            if option_parse.takes_val == TakesVal.YES and val is None:
                print(f"edulint: option {name} takes an argument but none was supplied", file=sys.stderr)
            elif option_parse.takes_val == TakesVal.NO and val is not None:
                print(f"edulint: option {name} takes no argument but {val} was supplied", file=sys.stderr)
            else:
                result.append(Arg(option, val))

    return result


def apply_translates(args: List[Arg], config_translations: Dict[Option, Translation]) -> Config:
    result: Config = Config(edulint=args)
    for arg in args:
        translated = config_translations.get(arg.option)
        if translated is not None:
            assert translated.to != Linters.EDULINT
            result.others[translated.to].extend(translated.val)
        elif arg.option == Option.PYLINT:
            assert arg.val
            result.others[Linters.PYLINT].append(arg.val)
        elif arg.option == Option.FLAKE8:
            assert arg.val
            result.others[Linters.FLAKE8].append(arg.val)
    return result


def get_config(
        filename: str, option_parses: Dict[Option, OptionParse] = get_option_parses(),
        config_translations: Dict[Option, Translation] = get_config_translations()) -> Config:
    return apply_translates(parse_args(extract_args(filename), option_parses), config_translations)



from enum import Enum


class Linters(Enum):
    EDULINT = 0,  # keep defined first
    PYLINT = 1,
    FLAKE8 = 2

    def __str__(self: Enum) -> str:
        return self.name.lower()

    @staticmethod
    def from_str(linter_str: str) -> "Linters":
        for linter in Linters:
            if str(linter) == linter_str.lower():
                return linter
        assert False, "no such linter: " + linter_str



from dataclasses import dataclass
from typing import Dict, List, Tuple
from enum import Enum, auto


class Option(Enum):
    PYLINT = auto()
    FLAKE8 = auto()
    ENHANCEMENT = auto()
    PYTHON_SPEC = auto()
    ALLOWED_ONECHAR_NAMES = auto()

    def to_name(self) -> str:
        return self.name.lower().replace("_", "-")

    @staticmethod
    def from_str(option_str: str) -> "Option":
        for option in Option:
            if option.to_name() == option_str.lower():
                return option
        assert False, "no such option: " + option_str


class TakesVal(Enum):
    YES = auto()
    NO = auto()
    OPTIONAL = auto()


@dataclass
class OptionParse:
    option: Option
    name: str
    help_: str
    takes_val: TakesVal


OPTIONS: List[Tuple[Option, str, TakesVal]] = [
    (
        Option.PYLINT,
        "arguments to be passed to pylint",
        TakesVal.YES
    ),
    (
        Option.FLAKE8,
        "arguments to be passed to flake8",
        TakesVal.YES
    ),
    (
        Option.ENHANCEMENT,
        "enable checking for ways to improve the code further",
        TakesVal.NO
    ),
    (
        Option.PYTHON_SPEC,
        "enable checking for ways to improve the code with Python-specific constructions",
        TakesVal.NO
    ),
    (
        Option.ALLOWED_ONECHAR_NAMES,
        "only listed characters are allowed to be variable names of length one",
        TakesVal.YES
    ),
]


def get_option_parses() -> Dict[Option, OptionParse]:
    return {option: OptionParse(option, option.to_name(), help_, takes_val) for option, help_, takes_val in OPTIONS}


def get_name_to_option(option_parses: Dict[Option, OptionParse]) -> Dict[str, Option]:
    return {parse.name: opt for opt, parse in option_parses.items()}



from edulint.linters import Linters
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.config.config import Config
from edulint.linting.problem import Problem
from dataclasses import dataclass
from typing import Callable, Optional, Pattern, Match, AnyStr, Dict, Tuple, Set, List
import re


@dataclass
class Tweaker:
    used_options: Set[Option]

    pattern: Pattern[AnyStr]  # type: ignore
    keep: Callable[["Tweaker", Problem, List[Arg]], bool]
    reword: Optional[Callable[["Tweaker", Problem], str]] = None

    def match(self, problem: Problem) -> Optional[Match[AnyStr]]:
        return self.pattern.match(problem.text)

    def should_keep(self, problem: Problem, args: List[Arg]) -> bool:
        return self.keep(self, problem, args)  # type: ignore

    def does_reword(self) -> bool:
        return self.reword is not None

    def get_reword(self, problem: Problem) -> str:
        return self.reword(self, problem) if self.reword else problem.text


def invalid_name_keep(self: Tweaker, problem: Problem, args: List[Arg]) -> bool:
    match = self.match(problem)
    assert match
    if match.group(1).lower() == "module":
        return False

    name = match.group(2)
    style = match.group(3)
    if len(name) == 1 and Config.has_opt_in(args, Option.ALLOWED_ONECHAR_NAMES):
        allowed_names = Config.get_val_from(args, Option.ALLOWED_ONECHAR_NAMES)
        assert allowed_names is not None
        return name not in allowed_names

    if style == "snake_case naming style" and any(ch1.islower() and ch2.isupper() for ch1, ch2 in zip(name, name[1:])):
        return True
    return False


Tweakers = Dict[Tuple[Linters, str], Tweaker]

TWEAKERS = {
    (Linters.PYLINT, "C0103"): Tweaker(  # invalid-name
        set([Option.PYTHON_SPEC, Option.ALLOWED_ONECHAR_NAMES]),
        re.compile(r"^(.*) name \"(.*)\" doesn't conform to (.*)$"),
        invalid_name_keep
    )
}


def get_tweakers() -> Tweakers:
    return TWEAKERS



from edulint.linting.problem import ProblemEncoder
from edulint.config.config import get_config
from edulint.linting.linting import lint
import argparse
import json


def setup_argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Lint provided code.")
    parser.add_argument("file", metavar="FILE", help="the file to lint")
    parser.add_argument("--json", action="store_true",
                        help="should output problems in json format")
    return parser.parse_args()


def main() -> int:
    args = setup_argparse()
    config = get_config(args.file)
    result = lint(args.file, config)
    if args.json:
        print(json.dumps(result, indent=1, cls=ProblemEncoder))
    else:
        for problem in result:
            print(problem)
    return 0



from typing import Dict


explanations: Dict[str, Dict[str, str]] = {
    "W0612": {
        "why": "Having unused variables in code makes the code unnecessarily complicated.",
        "examples": "A solution is to remove the variable, if possible, or rename it to `_`, which is a common"
        "convention for naming unused variables. A variable named `_` should never be used anywhere later in code."
    }
}


def get_explanations() -> Dict[str, Dict[str, str]]:
    return explanations
